import sys  # open stderr
import unicodedata  # case insensitive compare in Utility
from re import sub

import inflect  # singular and plural nouns
from requests import \
    Session  # HTTP user agent will not emit server cert warnings if verify=False
from urllib3.exceptions import InsecureRequestWarning

from requests import status_codes
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from urllib3 import disable_warnings

disable_warnings(InsecureRequestWarning)

class Utility:
    def __init__(self):
        pass

    def camel(self, snake_str):
        first, *others = snake_str.split('_')
        return ''.join([first.lower(), *map(str.title, others)])

    def snake(self, camel_str):
        return sub(r'(?<!^)(?=[A-Z])', '_', camel_str).lower()

    def normalize_caseless(self, text):
        return unicodedata.normalize("NFKD", text.casefold())

    def caseless_equal(self, left, right):
        return self.normalize_caseless(left) == self.normalize_caseless(right)

class LookupDict(dict):
    """Dictionary lookup object."""

    def __init__(self, name=None):
        self.name = name
        super(LookupDict, self).__init__()

    def __repr__(self):
        return '<lookup \'%s\'>' % (self.name)

    def __getitem__(self, key):
        # We allow fall-through here, so values default to None

        return self.__dict__.get(key, None)

    def get(self, key, default=None):
        return self.__dict__.get(key, default)

def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)

p = inflect.engine()
def plural(singular):
    # if already plural then return, else pluralize
    if singular[-1:] == 's':
        return(singular)
    else:
        return(p.plural_noun(singular))

def singular(plural):
    return(p.singular_noun(plural))

STATUSES_BY_CODE = {
    100: ('new', 'created'),
    200: ('building', 'incomplete', 'allocated'),
    300: ('active', 'complete', 'provisioned'),
    400: ('registered', 'enrolled'),
    500: ('error', 'server_error'),
    600: ('updating', 'modifying'),
    800: ('deleting', 'released', 'decommissioned'),
    900: ('defunct', 'deleted')
}

CODES = LookupDict(name='statuses')
for code, titles in STATUSES_BY_CODE.items():
    for title in titles:
        setattr(CODES, title.upper(), code)

RESOURCES = {
    'networks': {
        'embedded': "networkList",
        'create_responses': ["ACCEPTED"]
    },
    'endpoints': {
        'embedded': "endpointList",
        'create_responses': ["ACCEPTED"]
    },
    'edge-routers': {
        'embedded': "edgeRouterList",
        'create_responses': ["ACCEPTED"]
    },
    'edge-router-policies': {
        'embedded': "edgeRouterPolicyList",
        'create_responses': ["OK", "ACCEPTED"]
    },
    'app-wans': {
        'embedded': "appWanList",
        'create_responses': ["OK"]
    },
    'services': {
        'embedded': "serviceList",
        'create_responses': ["ACCEPTED"]
    },
    'service-policies': {
        'embedded': "servicePolicyList",
        'create_responses': ["ACCEPTED"]
    },
    'service-edge-router-policies': {
        'embedded': "serviceEdgeRouterPolicyList",
        'create_responses': ["ACCEPTED"]
    },
    'posture-checks': {
        'embedded': "postureCheckList",
        'create_responses': ["ACCEPTED"]
    }
}

# TODO: [MOP-13441] associate locations with a short list of major geographic regions / continents
MAJOR_REGIONS = {
    "AWS" : {
        "Americas": ("Canada Central","N. California","N. Virginia","Ohio","Oregon","Sao Paulo"),
        "EuropeMiddleEastAfrica": ("Bahrain","Cape Town South Africa","Frankfurt","Ireland","London","Milan","Paris","Stockholm"),
        "AsiaPacific": ("Hong Kong","Mumbai","Seoul","Singapore","Sydney","Tokyo")
    }
}

DC_PROVIDERS = ["AWS", "AZURE", "GCP", "OCP"]

HOST_PROPERTIES = ["ownerIdentityId", "ipAddress", "port", "provider", "providerInstanceId", "size", "locationMetadataId", "dataCenterId"]

EXCLUDED_PATCH_PROPERTIES = {
    "edge-routers": ["registration"],
    "services": [],
    "endpoints": [],
    "edge-router-policies": [],
    "networks": [],
    "app-wans": [],
    "posture-checks": []
}

VALID_SERVICE_PROTOCOLS = ["tcp", "udp"]
VALID_SEPARATORS = '[:-]' # : or - will match regex pattern

def docstring_parameters(*args, **kwargs):
    """Part a method's __doc__ string with format()."""
    def decorated(ref):
        ref.__doc__ = ref.__doc__.format(*args, **kwargs)
        return ref
    return decorated


RETRY_STRATEGY = Retry(
    total=3,
    status_forcelist=[413, 429, 503],
    method_whitelist=["HEAD", "GET", "PUT", "DELETE", "OPTIONS", "TRACE"],
    backoff_factor=1
)

DEFAULT_TIMEOUT = 31 # seconds, Gateway Service waits 30s before responding with an error code e.g. 503 and
# so waiting at least 31s is necessary to obtain that information
class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = DEFAULT_TIMEOUT
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)

http = Session()
# Mount it for both http and https usage
adapter = TimeoutHTTPAdapter(timeout=DEFAULT_TIMEOUT, max_retries=RETRY_STRATEGY)
http.mount("https://", adapter)
http.mount("http://", adapter)

STATUS_CODES = status_codes


import json
import os
import re  # regex
import time  # enforce a timeout; sleep
from pathlib import Path

import jwt  # decode the JWT claimset

from .utility import RESOURCES, STATUS_CODES, eprint, http


class Organization:
    """ Default is to use the Organization of the caller's user or API account identity
    :param: organization_id is optional string UUID of an alternative Organization
    :param: organization_label is optional string `label` property of an alternative Organization
    :param token: continue using a session with this optional token from an existing instance of Organization
    :param credentials: optional alternative path to API account credentials file, default is ~/.netfoundry/credentials.json
    :param proxy: optional HTTP proxy, e.g., http://localhost:8080
    """

    def __init__(self, 
        credentials=None, 
        organization_id: str=None, 
        organization_label: str=None,
        token=None, 
        proxy=None):

        # verify auth endpoint's server certificate if proxy is type SOCKS or None
        self.proxy = proxy
        if proxy is None:
            self.proxies = dict()
            self.verify = True
        else:
            self.proxies = {
                'http': proxy,
                'https': proxy
            }
            if proxy[0:5] == 'socks':
                self.verify = True
            else:
                self.verify = False
        
        # if not token then use standard env var if defined
        if token is not None:
            self.token = token
        elif 'NETFOUNDRY_API_TOKEN' in os.environ:
            self.token = os.environ['NETFOUNDRY_API_TOKEN']

        # if the token was found then extract the expiry
        try: 
            self.token
        except AttributeError: epoch = None
        else:
            claim = jwt.decode(jwt=self.token, algorithms=["RS256"], options={"verify_signature": False})
            # TODO: [MOP-13438] auto-renew token when near expiry (now+1hour in epoch seconds)
            expiry = claim['exp']
            epoch = time.time()

        client_id = None
        password = None
        token_endpoint = None

        # persist the credentials filename in instances so that it may be used to refresh the token
        if credentials is not None:
            self.credentials = credentials
            os.environ['NETFOUNDRY_API_ACCOUNT'] = self.credentials
        elif 'NETFOUNDRY_API_ACCOUNT' in os.environ:
            self.credentials = os.environ['NETFOUNDRY_API_ACCOUNT']
        # if any credentials var then require all credentials vars
        elif ('NETFOUNDRY_CLIENT_ID' in os.environ
                    or 'NETFOUNDRY_PASSWORD' in os.environ
                    or 'NETFOUNDRY_OAUTH_URL' in os.environ):
            if ('NETFOUNDRY_CLIENT_ID' in os.environ 
                    and 'NETFOUNDRY_PASSWORD' in os.environ 
                    and 'NETFOUNDRY_OAUTH_URL' in os.environ):
                client_id = os.environ['NETFOUNDRY_CLIENT_ID']
                password = os.environ['NETFOUNDRY_PASSWORD']
                token_endpoint = os.environ['NETFOUNDRY_OAUTH_URL']
            else:
                raise Exception("ERROR: some but not all credentials vars present. Need NETFOUNDRY_CLIENT_ID, NETFOUNDRY_PASSWORD, and NETFOUNDRY_OAUTH_URL or a credentials file in default file locations or NETFOUNDRY_API_ACCOUNT as path to credentials file.")
        else:
            self.credentials = "credentials.json"

        # if no token or near expiry (30 min) then use env vars or credentials file to obtain a token
        if epoch is None or epoch > (expiry - 600):
            # if not creds as env vars then look for creds file
            if not client_id and not password and not token_endpoint:
                # unless a valid path assume relative and search the default chain
                if not os.path.exists(self.credentials):
                    default_creds_chain = [
                        {
                            "scope": "project",
                            "base": str(Path.cwd())
                        },
                        {
                            "scope": "user",
                            "base": str(Path.home())+"/.netfoundry"
                        },
                        {
                            "scope": "device",
                            "base": "/netfoundry"
                        }
                    ]
                    for link in default_creds_chain:
                        candidate = link['base']+"/"+self.credentials
                        if os.path.exists(candidate):
                            print("INFO: using credentials in {path} (found in {scope}-default directory)".format(
                                scope=link['scope'],
                                path=candidate
                            ))
                            self.credentials = candidate
                            break
                else:
                    print("INFO: using credentials in {path}".format(
                        path=self.credentials
                    ))

                try:
                    with open(self.credentials) as file:
                        try: account = json.load(file)
                        except: raise Exception("ERROR: failed to load JSON from {file}".format(file=file))
                except: raise Exception("ERROR: failed to open {file} while working in {dir}".format(
                    file=self.credentials,dir=str(Path.cwd())))
                token_endpoint = account['authenticationUrl']
                client_id = account['clientId']
                password = account['password']

            # extract the environment name from the authorization URL aka token API endpoint
            self.environment = re.sub(r'https://netfoundry-([^-]+)-.*', r'\1', token_endpoint, re.IGNORECASE)
            # re: scope: we're not using scopes with Cognito, but a non-empty value is required;
            #  hence "/ignore-scope"
            scope = "https://gateway."+self.environment+".netfoundry.io//ignore-scope"
            # we can gather the URL of the API from the first part of the scope string by
            #  dropping the scope suffix
            self.audience = scope.replace('/ignore-scope','')
            # e.g. https://gateway.production.netfoundry.io/
            assertion = {
                "scope": scope,
                "grant_type": "client_credentials"
            }
            # request a token
            try:
                response = http.post(
                    token_endpoint,
                    auth=(client_id, password),
                    data=assertion,
                    verify=self.verify,
                    proxies=self.proxies)
                response_code = response.status_code
            except:
                eprint(
                    'ERROR: failed to contact the authentication endpoint: {}'.format(token_endpoint)
                )
                raise

            if response_code == STATUS_CODES.codes.OK:
                try:
                    token_text = json.loads(response.text)
                    self.token = token_text['access_token']
                except:
                    raise Exception(
                        'ERROR: failed to find an access_token in the response and instead got: {}'.format(
                            response.text
                        )
                    )
            else:
                raise Exception(
                    'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                        STATUS_CODES._codes[response_code][0].upper(),
                        response_code,
                        response.text
                    )
                )

        # learn about the environment from the token
        try:
            claim = jwt.decode(jwt=self.token, algorithms=["RS256"], options={"verify_signature": False})
            iss = claim['iss']
            if re.match(r'https://cognito-', iss):
                self.environment = re.sub(r'https://gateway\.([^.]+)\.netfoundry\.io.*',r'\1',claim['scope'])
            elif re.match(r'.*\.auth0\.com', iss):
                self.environment = re.sub(r'https://netfoundry-([^.]+)\.auth0\.com.*',r'\1',claim['iss'])
            self.audience = 'https://gateway.'+self.environment+'.netfoundry.io/'
        except: raise

        # always resolve Network Groups so we can specify either name or ID when calling super()
        self.network_groups = self.get_network_groups_by_organization()
        # Obtain a session token and find own `.caller` identity and `.organizations`
        self.caller = self.get_caller_identity()
        if organization_id:
            self.describe = self.get_organization(id=organization_id)
        elif organization_label:
            self.organizations_by_label = dict()
            for org in self.get_organizations():
                self.organizations_by_label[org['label']] = org['id']
            if organization_label in self.organizations_by_label.keys():
                self.describe = self.get_organization(id=self.organizations_by_label[organization_label])
            else:
                raise Exception(
                    'ERROR: failed to find org label {:s} in the list of orgs {:s}'.format(
                        organization_label,
                        str(self.organizations_by_label.keys())
                    )
                )

        else:
            self.describe = self.get_organization(id=self.caller['organizationId'])

        self.label = self.describe['label']
        self.id = self.describe['id']

        self.network_groups_by_name = dict()
        for group in self.network_groups:
            self.network_groups_by_name[group['name']] = group['id']

    def network_groups(self):
        return(self.get_network_groups_by_organization())
        
    def get_caller_identity(self):
        """return the caller's identity object
        """
        # try the API account endpoint first, then the endpoint for human, interactive users
        request = {
            "url": self.audience+'identity/v1/api-account-identities/self',
            "proxies": self.proxies,
            "verify": self.verify,
            "headers": { "authorization": "Bearer " + self.token }
        }
        try:
            response = http.get(**request)
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                caller = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting caller\'s API account identity from response document')
                raise(e)
        else:
            try:
                request["url"] = self.audience+'identity/v1/user-identities/self'
                response = http.get(**request)
                response_code = response.status_code
            except:
                raise

            if response_code == STATUS_CODES.codes.OK: # HTTP 200
                try:
                    caller = json.loads(response.text)
                except ValueError as e:
                    eprint('ERROR getting caller\'s user identity from response document')
                    raise(e)
            else:
                raise Exception(
                    'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                        STATUS_CODES._codes[response_code][0].upper(),
                        response_code,
                        response.text
                    )
                )

        return(caller)

    def get_organizations(self):
        """return the list of Organizations (formerly "tenants")
        """
        try:
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'identity/v1/organizations',
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                organizations = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting Network Groups')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(organizations)

    def get_organization(self, id):
        """return a single Organizations by ID
        """
        try:
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'identity/v1/organizations/'+id,
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                organization = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting Network Groups')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(organization)

    def get_network_group(self,network_group_id):
        """describe a Network Group
        """
        try:
            # /network-groups/{id} returns a Network Group object
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'rest/v1/network-groups/'+network_group_id,
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                network_group = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting Network Group')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(network_group)

    def get_network(self,network_id):
        """describe a Network by ID
        """
        try:
            # /networks/{id} returns a Network object
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'rest/v1/networks/'+network_id,
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                network = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting Network Group')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(network)

    def get_network_groups_by_organization(self):
        """list Network Groups
        TODO: filter by Organization when that capability is available
        """
        try:
            # /network-groups returns a list of dicts (Network Group objects)
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'rest/v1/network-groups',
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                network_groups = json.loads(response.text)['_embedded']['organizations']
            except ValueError as e:
                eprint('ERROR getting Network Groups')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(network_groups)

    def get_networks_by_organization(self):
        """
        return all networks known to this Organization
        """

        try:
            # returns a list of dicts (network objects)
            headers = { "authorization": "Bearer " + self.token }
            response = http.get(
                self.audience+'core/v2/networks',
                proxies=self.proxies,
                verify=self.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                networks = json.loads(response.text)['_embedded'][RESOURCES['networks']['embedded']]
            except KeyError:
                networks = []
                pass
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(networks)

    def get_networks_by_group(self,network_group_id):
        """Find networks by network group ID.

        :param network_group_id: required network group UUIDv4
        """
        try:
            headers = { 
                "authorization": "Bearer " + self.token 
            }
            params = {
                "findByNetworkGroupId": network_group_id
            }
            response = http.get(
                self.audience+'core/v2/networks',
                proxies=self.proxies,
                verify=self.verify,
                headers=headers,
                params=params
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                embedded = json.loads(response.text)
                networks = embedded['_embedded'][RESOURCES['networks']['embedded']]
            except KeyError:
                networks = []
                pass
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(networks)



def waitForHttpResponse(self, status, type, netId, wait, id=0):
    """continuously poll for the expected HTTP response code until expiry
    :param type: the type of entity e.g. network, endpoint, service
    :param id: the UUID of the entity having a status
    :param status: string that is the symbolic name of the expected status code
    :param wait: SECONDS after which to raise an exception
    Example: status NOT_FOUND will succeed when response is HTTP 404
    """

    now = time.time()

    # poll for status until expiry
    sys.stdout.write(
        '\twaiting for status {:s} ({:d}) or until {:s}.'.format(
            status,
            requests.status_codes.codes[status],
            time.ctime(now+wait)
        )
    )
    # resolve the expected status to an HTTP code
    status_code = requests.status_codes.codes[status]
    # initialize a variable to store the last checked HTTP code
    http_code = int()
    while time.time() < now+wait and not status_code == http_code:
        sys.stdout.write('.')
        sys.stdout.flush()
        try:
            entity = self.getResourceStatus(type=type,
                                    netId=netId,
                                    id=id)
            if not http_code or (
                http_code and
                not http_code == entity['http_code']
            ):
                sys.stdout.write(
                    '\n\t\t:'
                    +'{:^19s}'.format(
                        requests.status_codes._codes[
                            entity['http_code']
                        ][0].upper()
                    )
                    +':'

                )

            http_code = entity['http_code']
        except:
            raise

        time.sleep(15)

    print() # newline terminates progress meter

    if status_code == http_code:
        return(True)
    else:
        raise Exception(
            'timed out with status {} ({}) while waiting for {} ({})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code,
                status,
                status_code,
            )
        )

def patchAttributes(token,type,resource):
    """return the new resource object
        :token [required] the API access token
        :type [required] one of endpoints, edge-routers, services
        :resource [required] the resource object to PUT 
    """
    try:
        headers = { 
            'Content-Type': 'application/json',
            "authorization": "Bearer " + token 
        }
        response = requests.patch(
            self.audience+'/core/v2/'+type+'/'+resource['id'],
            json={
                'name': resource['name'],
                'attributes': resource['attributes']
            },
            headers=headers
        )
        http_code = response.status_code
    except:
        raise

    if http_code == requests.status_codes.codes[RESOURCES[type]['expect']]: # HTTP 200
        try:
            updated = json.loads(response.text)
        except ValueError as e:
            eprint('ERROR: failed to load updated resource object from PUT response body')
            raise(e)
    else:
        raise Exception(
            'unexpected response: {} (HTTP {:d}\n{})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code,
                response.text
            )
        )

    return(updated)

def post(self,
    method,
    json,
    expected_response="ACCEPTED"):
    """
    post arbitrary data as JSON to an API method
    """

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(
            self.audience+"core/v2"+method,
            json=json,
            headers=headers,
            proxies=self.proxies,
            verify=self.verify
        )
        http_code = response.status_code
    except:
        raise

    if not http_code == requests.status_codes.codes[expected_response]:
        raise Exception('unexpected response: {} (HTTP {:d})'.format(
            requests.status_codes._codes[http_code][0].upper(),
            http_code)
        )
    text = response.text
    return(text)


def createIpNetworkService(
        self,
        netId,
        name,
        gatewayIp,
        gatewayNetmask,
        interceptIp,
        endpointId,
        wait=0):
    """publish a subnetwork (IP or range described by network address and netmask)
    :param netId: network UUID
    :param name: descriptive name allowing whitespace
    :param gatewayIp: the network address of the subnet (lowest IP)
    :param gatewayNetmask: the netmask describing the size of the published subnet
    :param interceptIp: the network address of the subnet used by clients
    :param endpointId: UUID of the serving gateway that will provide this service
    :param wait: optional wait seconds for service to become ACTIVE
    """
    request = {
        "name": name,
        "serviceClass": "GW",
        "gatewayIp": gatewayIp,
        "gatewayNetmask": gatewayNetmask,
        "interceptIp": interceptIp,
        "endpointId": endpointId
    }

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(
            self.audience+"core/v2/networks/"+netId+"/services",
            json=request,
            headers=headers,
            proxies=self.proxies,
            verify=self.verify)
        http_code = response.status_code
    except:
        raise

    if not http_code == requests.status_codes.codes.ACCEPTED:
        raise Exception(
            'unexpected response: {} (HTTP {:d})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code
            )
        )

    svcId = json.loads(response.text)['_links']['self']['href'].split('/')[-1]

    # expected value is UUID of new endpoint
    UUID(svcId, version=4) # validate the returned value is a UUID

    if not wait == 0:
        try:
            self.waitForStatus(
                status='ACTIVE',
                type='service',
                netId=netId,
                id=svcId,
                wait=wait)
        except:
            raise

    return(svcId)

def createAwsGateway(self, name, netId, dataCenterId, wait=0, family="dvn", managed=False):
    """
    create an AWS gateway endpoint with
    :param name: gateway name
    :param netId: network UUID
    :param dataCenterId: datacenter UUID
    :param wait: optional wait seconds for endpoint to become REGISTERED (400)
    :param family: optional family indicating endpoint type if not "dvn"
    :param managed: optional boolean indicating instance is launched and managed by MOP
    """

    if not managed and family == "ziti":
        endpointType = "ZTNHGW"
    elif not managed and family == "dvn":
        endpointType = "AWSCPEGW"
    elif managed and family == "ziti":
        endpointType = "ZTGW"
    elif managed and family == "dvn":
        endpointType = "GW"
    else:
        raise Exception(
            'unexpected family "{}" or endpoint type "{}"'.format(
                family,
                endpointType
            )
        )

    request = {
        "name": name,
        "endpointType": endpointType,
        "dataCenterId": dataCenterId
    }

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(self.audience+"core/v2/networks/"+netId+"/endpoints",
                                    json=request,
                                    headers=headers,
                                    proxies=self.proxies,
                                    verify=self.verify
                                )
        http_code = response.status_code
    except:
        raise

    if not http_code == requests.status_codes.codes.ACCEPTED:
        raise Exception(
            'unexpected response: {} (HTTP {:d})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code
            )
        )

    endId = json.loads(response.text)['_links']['self']['href'].split('/')[-1]

    # expected value is UUID of new endpoint
    UUID(endId, version=4) # validate the returned value is a UUID

    if not wait == 0:
        try:
            self.waitForStatus(status='REGISTERED',
                                        type='endpoint',
                                        netId=netId,
                                        id=endId,
                                        wait=wait)
        except:
            raise

    return(endId)

def createVcpeGateway(self, name, netId, geoRegionId, wait=0):
    """
    create a self-hosted gateway endpoint with
    :param name: gateway name
    :param netId: network UUID
    :param geoRegionId: geo region UUID
    :param wait: optional wait seconds for endpoint to become REGISTERED (400)
    """
    request = {
        "name": name,
        "endpointType": "VCPEGW",
        "geoRegionId": geoRegionId
    }

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(self.audience+"core/v2/networks/"+netId+"/endpoints",
                                    json=request,
                                    headers=headers,
                                    proxies=self.proxies,
                                    verify=self.verify
                                )
        http_code = response.status_code
    except:
        raise

    if not http_code == requests.status_codes.codes.ACCEPTED:
        raise Exception(
            'unexpected response: {} (HTTP {:d})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code
            )
        )

    endId = json.loads(response.text)['_links']['self']['href'].split('/')[-1]

    # expected value is UUID of new endpoint
    UUID(endId, version=4) # validate the returned value is a UUID

    if not wait == 0:
        try:
            self.waitForStatus(status='REGISTERED',
                                        type='endpoint',
                                        netId=netId,
                                        id=endId,
                                        wait=wait)
        except:
            raise

    return(endId)

def createIpHostService(self,
                        netId,
                        name,
                        networkIp,
                        networkFirstPort,
                        networkLastPort,
                        interceptIp,
                        interceptFirstPort,
                        interceptLastPort,
                        protocolType,
                        endpointId,
                        wait=0
                        ):
    """publish a host:port(s) couplet
    :param netId: network UUID
    :param name: descriptive name allowing whitespace
    :param networkIp: interface or masquerade address of the published service
    :param networkFirstPort: lower bound of published port range
    :param networkLastPort: upper bound of published port range
    :param interceptIp: destination address used by clients
    :param interceptFirstPort: lower bound of destination port range
    :param interceptLastPort: upper bound of destination port range
    :param protocolType: one of tcp, udp
    :param endpointId: UUID of the serving gateway that provides this service
    :param wait: optional wait seconds for service to become ACTIVE
    """
    request = {
        "name": name,
        "serviceClass": "CS",
        "networkIp": networkIp,
        "networkFirstPort": networkFirstPort,
        "networkLastPort": networkLastPort,
        "interceptIp": interceptIp,
        "interceptFirstPort": interceptFirstPort,
        "interceptLastPort": interceptLastPort,
        "protocolType": protocolType,
        "endpointId": endpointId,
        "serviceType": protocolType.upper(),
        "transparency": "NO",
        "bridgeStp": "YES",
        "collectionLocation": "BOTH",
        "cryptoLevel": "STRONG",
        "dnsOptions": "NONE",
        "icmpTunnel": "YES",
        "localNetworkGateway": "YES",
        "multicast": "OFF",
        "pbrType": "WAN",
        "permanentConnection": "NO",
        "rateSmoothing": "YES",
        "serviceInterceptType": "IP"
    }
    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(self.audience+"core/v2/networks/"+netId+"/services",
                                    json=request,
                                    headers=headers,
                                    proxies=self.proxies,
                                    verify=self.verify
                                )
        http_code = response.status_code
    except:
        raise

    if not http_code == requests.status_codes.codes.ACCEPTED:
        raise Exception(
            'unexpected response: {} (HTTP {:d})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code
            )
        )

    svcId = json.loads(response.text)['_links']['self']['href'].split('/')[-1]

    # expected value is UUID of new endpoint
    UUID(svcId, version=4) # validate the returned value is a UUID

    if not wait == 0:
        try:
            self.waitForStatus(status='ACTIVE',
                                        type='service',
                                        netId=netId,
                                        id=svcId,
                                        wait=wait)
        except:
            raise

    return(svcId)

def createAppWan(self,
                    netId,
                    name,
                    wait=0
                ):
    """authorize endpoint(s) for service(s) where
    :param netId: network UUID
    :param name: descriptive name allowing whitespace
    """
    request = {
        "name": name,
    }

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    try:
        response = requests.post(self.audience+"core/v2/networks/"+netId+"/appWans",
                                    json=request,
                                    headers=headers,
                                    proxies=self.proxies,
                                    verify=self.verify
                                )
        #print(response)
        http_code = response.status_code
    except:
        raise

    if http_code == requests.status_codes.codes.ACCEPTED:
        appWanId = json.loads(response.text)['_links']['self']['href'].split('/')[-1]
    else:
        raise Exception(
            'unexpected response: {} (HTTP {:d})'.format(
                requests.status_codes._codes[http_code][0].upper(),
                http_code
            )
        )

    if not wait == 0:
        try:
            self.waitForStatus(status='ACTIVE',
                                        type='appWan',
                                        netId=netId,
                                        id=appWanId,
                                        wait=wait)
        except:
            raise

    return(appWanId)

def updateAppWan(self,
                    appWanId,
                    netId,
                    services,
                    endpoints=[],
                    endpointGroups=[],
                    wait=0
                ):
    """authorize endpoint(s) or endpoint group(s) or both for service(s) where
    :param appWanId: AppWAN UUID to update
    :param netId: network UUID housing the AppWAN
    :param services: list of service UUIDs to publish
    :param endpoints: optional list of endpoint UUIDs to authorize
    :param endpointGroups: optional list of endpoint group UUIDs to authorize
    """

    appWan = dict()

    appWan['services'] = {
        "ids": services
    }

    appWan['endpoints'] = {
        "ids": endpoints
    }

    appWan['endpointGroups'] = {
        "ids": endpointGroups
    }

    headers = {
        'Content-Type': 'application/json',
        "authorization": "Bearer " + self.token
    }

    for resource in ['services', 'endpoints', 'endpointGroups']:
        try:
            response = requests.post(
                self.audience+"core/v2/networks/"+netId+"/appWans/"+appWanId+"/"+resource,
                        json=appWan[resource],
                        headers=headers,
                        proxies=self.proxies,
                        verify=self.verify)
            #print(response)
            http_code = response.status_code
        except:
            raise

        if not http_code == requests.status_codes.codes.ACCEPTED:
            raise Exception(
                'unexpected response: {} (HTTP {:d})'.format(
                    requests.status_codes._codes[http_code][0].upper(),
                    http_code
                )
            )

        if not wait == 0:
            try:
                self.waitForStatus(status='ACTIVE',
                                            type='appWan',
                                            netId=netId,
                                            id=appWanId,
                                            wait=wait)
            except:
                raise

    return(True)

def describeResource(self, type, netId, id=0):
    """return the full object describing an entity
    :param type: the type of entity e.g. network, endpoint, service
    :param netId: the UUID of the network housing the entity
    :param id: the UUID of the entity having a status if not a network
    """

    try:
        headers = { "authorization": "Bearer " + self.token }
        entityUrl = self.audience+'core/v2/networks/'+netId
        if not type == 'network':
            if id == 0:
                raise Exception("ERROR: entity UUID must be specified if not a network")
            entityUrl += '/'+type+'s/'+id

        response = requests.get(entityUrl,
                                proxies=self.proxies,
                                verify=self.verify,
                                headers=headers)
        http_code = response.status_code
    except:
        raise

    if http_code == requests.status_codes.codes.OK:
        try:
            entity = json.loads(response.text)
        except:
            eprint('ERROR parsing entity object in response')
            raise
    else:
        status = None

    return(entity)


#!/usr/bin/env python3
r"""Command-line interface to the NetFoundry API
Usage::
    $ python3 -m netfoundry.ctl --credentials credentials.json

PYTHON_ARGCOMPLETE_OK
"""
import pkg_resources  # part of setuptools
from milc import cli

import netfoundry


@cli.argument('-c', '--credentials', help='API account JSON file from web console', default='credentials.json')
@cli.argument('-p', '--proxy', help='HTTP proxy URL e.g. http://locahost:8888', default=None)
@cli.subcommand('version', help='print Python module version')
@cli.entrypoint('Greet a user.')

def main(cli):
    cli.log.info('Hello, %s!', cli.config.general.credentials)

    package = "netfoundry"
    version = pkg_resources.require(package)[0].version

    session = netfoundry.Session(
        credentials=cli.config.general.credentials if cli.config.general.credentials else None,
        proxy=cli.config.general.proxy
    )

    # yields a list of Network Groups in Organization.network_groups[], but there's typically only one group
    Organization = netfoundry.Organization(session)

    # use the default Network Group (the first Network Group ID known to the Organization)
    network_group = netfoundry.NetworkGroup(Organization)

if __name__ == '__main__':
    cli()


import json

from .utility import RESOURCES, STATUS_CODES, eprint, http


class NetworkGroup:
    """use a Network Group by name or ID.
    
    The default is to use the first network group available to the organization of the caller.
    """

    def __init__(self, Organization: object, network_group_id: str=None, network_group_name: str=None):
        """Initialize the network group class with a group name or ID."""
        if network_group_id:
            self.network_group_id = network_group_id
            self.network_group_name = [ ng['organizationShortName'] for ng in Organization.network_groups if ng['id'] == network_group_id ][0]
        # TODO: review the use of org short name ref https://netfoundry.slack.com/archives/C45UDKR8V/p1603655594135000?thread_ts=1580318187.149400&cid=C45UDKR8V
        elif network_group_name:
            self.network_group_name = network_group_name
            network_group_matches = [ ng['id'] for ng in Organization.network_groups if ng['organizationShortName'] == network_group_name ]
            if len(network_group_matches) == 1:
                self.network_group_id = [ ng['id'] for ng in Organization.network_groups if ng['organizationShortName'] == network_group_name ][0]
            else:
                raise Exception("ERROR: there was not exactly one network group matching the name \"{}\"".format(network_group_name))
        elif len(Organization.network_groups) > 0:
            # first Network Group is typically the only Network Group
            self.network_group_id = Organization.network_groups[0]['id']
            self.network_group_name = Organization.network_groups[0]['organizationShortName']
            # warn if there are other groups
            if len(Organization.network_groups) > 1:
                eprint("WARN: Using first Network Group {:s} and ignoring {:d} other(s) e.g. {:s}, etc...".format(
                    self.network_group_name,
                    len(Organization.network_groups) - 1,
                    Organization.network_groups[1]['organizationShortName']
                ))
            elif len(Organization.network_groups) == 1:
                eprint("WARN: Using the default Network Group: {:s}".format(
                    self.network_group_name
                ))
        else:
            raise Exception("ERROR: need at least one Network Group in organization")

        self.session = Organization
        self.describe = Organization.get_network_group(self.network_group_id)
        self.id = self.network_group_id
        self.name = self.network_group_name
        self.vanity = Organization.label.lower()

        if self.session.environment == "production":
            self.nfconsole = "https://{vanity}.nfconsole.io".format(vanity=self.vanity)
        else:
            self.nfconsole = "https://{vanity}.{env}-nfconsole.io".format(vanity=self.vanity, env=self.session.environment)

    def nc_data_centers(self): # this was attribute self.nc_data_centers and converted to method to avoid calling preemptively with self.__init__
        return(self.get_controller_data_centers())

    def nc_data_centers_by_location(self): # this was attribute self.nc_data_centers_by_location and converted to method to avoid calling preemptively with self.__init__
        my_nc_data_centers_by_location = dict()
        for dc in self.get_controller_data_centers():
            my_nc_data_centers_by_location[dc['locationCode']] = dc['id']
            # e.g. { us-east-1: 02f0eb51-fb7a-4d2e-8463-32bd9f6fa4d7 }
        return(my_nc_data_centers_by_location)

    # resolve network UUIDs by name
    def networks_by_name(self): # this was attribute self.networks_by_name and converted to method to avoid calling preemptively with self.__init__
        my_networks_by_name = dict()
        for net in self.session.get_networks_by_group(self.network_group_id):
            my_networks_by_name[net['name']] = net['id']
        return(my_networks_by_name)

    def get_controller_data_centers(self):
        """list the data centers where a Network Controller may be created
        """
        try:
            # data centers returns a list of dicts (data center objects)
            headers = { "authorization": "Bearer " + self.session.token }
            params = {
                # "productVersion": self.product_version,
                # "hostType": "NC",
                # "provider": "AWS"
            }
            response = http.get(
                self.session.audience+'core/v2/data-centers',
                proxies=self.session.proxies,
                verify=self.session.verify,
                headers=headers,
                params=params
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                all_data_centers = json.loads(response.text)['_embedded']['dataCenters']
                aws_data_centers = [dc for dc in all_data_centers if dc['provider'] == "AWS"]
            except ValueError as e:
                eprint('ERROR getting data centers')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        return(aws_data_centers)

    def get_product_metadata(self, is_active: bool=True):
        """fetch all product metadata
        """
        try:
            response = http.get(
                self.session.audience+'product-metadata/v2/download-urls.json',
                proxies=self.session.proxies,
                verify=self.session.verify
            )
            response_code = response.status_code
        except:
            raise

        if response_code == STATUS_CODES.codes.OK: # HTTP 200
            try:
                product_metadata = json.loads(response.text)
            except ValueError as e:
                eprint('ERROR getting product metadata')
                raise(e)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        if is_active:
            active_product_metadata = dict()
            for product in product_metadata.keys():
                if product_metadata[product]['active']:
                    active_product_metadata[product] = product_metadata[product]
            return (active_product_metadata)
        else:
            return (product_metadata)

    def list_product_versions(self, product_metadata: dict={}):
        """find all product version from product metadata
        """
        if product_metadata:
            product_versions = product_metadata.keys()
        else:
            product_metadata = self.get_product_metadata()
            product_versions = product_metadata.keys()

        return (product_versions)

    def find_latest_product_version(self, product_versions: list=[]):
        """find the highest sorted product version number (may be experimental, not stable)
        """
        if not product_versions:
            product_versions = self.list_product_versions()

        from distutils.version import LooseVersion
        return sorted(product_versions, key=LooseVersion)[-1]

    def create_network(self, name: str, network_group_id: str=None, location: str="us-east-1", version: str=None, size: str="small"):
        """
        create a network with
        :param name: required network name
        :param network_group: optional Network Group ID
        :param location: optional data center region name in which to create
        :param version: optional product version string like 7.3.17
        :param size: optional network configuration metadata name from /core/v2/network-configs e.g. "medium"
        """
        
        my_nc_data_centers_by_location = self.nc_data_centers_by_location()
        if not location in my_nc_data_centers_by_location.keys():
            raise Exception("ERROR: unexpected Network location '{:s}'. Valid locations include: {}.".format(location, my_nc_data_centers_by_location.keys()))

        request = {
            "name": name,
            "locationCode": location,
            "size": size,
        }
        if network_group_id:
            request["networkGroupId"] = network_group_id
        else:
            request["networkGroupId"] = self.network_group_id

        if version:
            product_versions = self.list_product_versions()
            if version == "latest":
                request['productVersion'] = self.find_latest_product_version(product_versions)
            elif version in product_versions:
                request['productVersion'] = version
            elif version == "default":
                pass # do not specify a value for productVersion
            else:
                raise Exception("ERROR: invalid version \"{version}\". Expected one of {product_versions}".format(
                    version=version,
                    product_versions=product_versions))

        headers = {
            'Content-Type': 'application/json',
            "authorization": "Bearer " + self.session.token
        }

        try:
            response = http.post(
                self.session.audience+"core/v2/networks",
                proxies=self.session.proxies,
                verify=self.session.verify,
                json=request,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        any_in = lambda a, b: any(i in b for i in a)
        response_code_symbols = [s.upper() for s in STATUS_CODES._codes[response_code]]
        if any_in(response_code_symbols, RESOURCES['networks']['create_responses']):
            try:
                network = json.loads(response.text)
            except ValueError as e:
                raise e('ERROR: failed to load created network JSON, got HTTP code {:s} ({:d}) with body {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text)
                )
            else:
                return(network)
        else:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

    def delete_network(self, network_id=None, network_name=None):
        """
        delete a Network
        :param id: optional Network UUID to delete
        :param name: optional Network name to delete
        """
        try:
            networks_by_name = self.networks_by_name()
            if network_id:
#                import epdb; epdb.serve()
                network_name = next(name for name, uuid in networks_by_name.items() if uuid == network_id)
            elif network_name and network_name in networks_by_name.keys():
                network_id = networks_by_name[network_name]
        except:
            raise Exception("ERROR: need one of network_id or network_name for a Network in this Network Group: {:s}".format(self.name))

        try:
            headers = { "authorization": "Bearer " + self.session.token }
            entity_url = self.session.audience+'core/v2/networks/'+network_id
            response = http.delete(
                entity_url,
                proxies=self.session.proxies,
                verify=self.session.verify,
                headers=headers
            )
            response_code = response.status_code
        except:
            raise

        if not response_code == STATUS_CODES.codes.ACCEPTED:
            raise Exception(
                'ERROR: got unexpected HTTP code {:s} ({:d}) and response {:s}'.format(
                    STATUS_CODES._codes[response_code][0].upper(),
                    response_code,
                    response.text
                )
            )

        network = json.loads(response.text)
        return(network)




#!/usr/bin/env python3
r"""Use this script to create a network with functioning services and demonstrate the NetFoundry Python module.

Usage::
    $ python3 -m netfoundry.demo --network BibbidiBobbidiBoo
"""

import argparse
import os
import random
import sys
from pathlib import Path

import netfoundry
from netfoundry.utility import DC_PROVIDERS


def main():
    """Run the demo script."""
    print("DEBUG: running demo script in \"{:s}\"".format(sys.argv[0]))

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "command",
        nargs='?',
        default="create",
        help="the command to run",
        choices=["create","delete"]
    )
    parser.add_argument(
        "-y", "--yes",
        dest="yes",
        default=False,
        action="store_true",
        help="Skip interactive prompt to confirm destructive actions."
    )
    parser.add_argument(
        "-n", "--network",
        help="The name of your demo network"
    )
    parser.add_argument(
        "-o", "--organization",
        help="The label of an alternative organization (default is Org of caller)"
    )
    parser.add_argument(
        "-g", "--network-group",
        dest="network_group",
        help="The shortname of a Network Group (default is the first, typically singular, Group known to this Org)"
    )
    parser.add_argument(
        "-s", "--network-size",
        dest="size",
        default="small",
        help="Network size to create",
        choices=["small","medium","large"]
    )
    parser.add_argument(
        "-v", "--network-version",
        dest="version",
        default="default",
        help="Network product version: \"default\", \"latest\", or semver"
    )
    parser.add_argument(
        "-p", "--create-private",
        dest="private",
        default=False,
        action="store_true",
        help="Also create private Endpoint-hosted Services for the optional Docker Compose portion of the quickstart"
    )
    parser.add_argument(
        "-c", "--create-client",
        dest="client",
        default=False,
        action="store_true",
        help="Also create a client Endpoint for the optional Linux portion of the quickstart"
    )
    parser.add_argument("--credentials",
        default=None,
        help="path to API account credentials JSON file overrides NETFOUNDRY_API_ACCOUNT"
    )
    parser.add_argument("--provider",
        default="AWS",
        required=False,
        help="cloud provider to host Edge Routers",
        choices=DC_PROVIDERS
    )
    regions_group = parser.add_mutually_exclusive_group(required=False)
    regions_group.add_argument("--regions",
        default=["Americas"],
        nargs="+",
        help="space-separated one or more major geographic regions in which to place Edge Routers for overlay fabric",
        choices=["Americas", "EuropeMiddleEastAfrica", "AsiaPacific"]
    )
    regions_group.add_argument("--location-codes",
        dest="location_codes",
        nargs="+",
        help="cloud location codes in which to host Edge Routers"
    )
    parser.add_argument("--proxy",
        default=None,
        help="'http://localhost:8080'"+
        " 'socks5://localhost:9046'"
    )
    args = parser.parse_args()

    network_name = args.network
    
    # use the session with some organization, default is to use the first and there's typically only one
    organization = netfoundry.Organization(
        credentials=args.credentials if 'credentials' in args else None,
        organization_label=args.organization if 'organization' in args else None,
        proxy=args.proxy
    )

    # use some Network Group, default is to use the first and there's typically only one
    network_group = netfoundry.NetworkGroup(
        organization,
        network_group_name=args.network_group if 'network_group' in args else None
    )

    # create a Network
    if network_name in network_group.networks_by_name().keys():
        # use the Network
        network = netfoundry.Network(network_group, network_name=network_name)
        if args.command == "create":
            network.wait_for_status("PROVISIONED",wait=999,progress=True)
        elif args.command == "delete":
            if args.yes or query_yes_no("Permanently destroy Network \"{network_name}\" now?".format(network_name=network_name)):
                network.delete_network(progress=True)
            else:
                print("Not deleting Network \"{network_name}\".".format(network_name=network_name))
            sys.exit()
    elif args.command == "create":
        network_id = network_group.create_network(name=network_name,size=args.size,version=args.version)['id']
        network = netfoundry.Network(network_group, network_id=network_id)
        network.wait_for_status("PROVISIONED",wait=999,progress=True)
    elif args.command == "delete":
        print("Network \"{network_name}\" does not exist.".format(network_name=network_name))
        sys.exit()
    else:
        raise Exception("ERROR: failed to find a network named \"{:s}\"".format(network_name))

    # existing hosted routers
    hosted_edge_routers = network.edge_routers(only_hosted=True)
    # a list of places where Endpoints are dialing from

    # a list of locations to place a hosted router
    fabric_placements = list()
    if args.location_codes:
        for location in args.location_codes:
            existing_count = len([er for er in hosted_edge_routers if er['provider'] == args.provider and er['locationCode'] == args.location_code])
            if existing_count < 1:
                fabric_placements += [location]
            else:
                print("INFO: found a hosted Edge Router(s) in {location}".format(location=location))

        for location in fabric_placements:
            er = network.create_edge_router(
                name=location['locationName']+" ["+location['provider']+"]",
                attributes=[
                    "#defaultRouters",
                    "#"+location['locationCode'],
                    "#"+location['provider']
                ],
                provider=args.provider,
                location_code=location['locationCode']
            )
            hosted_edge_routers += [er]
            print("INFO: Placed Edge Router in {provider} ({location_name})".format(
                provider=location['provider'],
                location_name=location['locationName']
            ))
    elif args.regions:
        if args.provider or args.location_codes:
            print("WARN: ignoring provider and location_codes because AWS georegion is specified.")
        geo_regions = args.regions
        for region in geo_regions:
            data_center_ids = [dc['id'] for dc in network.aws_geo_regions[region]]
            existing_count = len([er for er in hosted_edge_routers if er['dataCenterId'] in data_center_ids])
            if existing_count < 1:
                choice = random.choice(network.aws_geo_regions[region]) # nosec
                # append the current major region to the randomly-chosen data center object
                #   so we can use it as a role attribute when we create the hosted Edge Router
                choice['geoRegion'] = region
                fabric_placements += [choice]
            else:
                print("INFO: found a hosted Edge Router(s) in {major}".format(major=region))

        for location in fabric_placements:
            er = network.create_edge_router(
                name=location['locationName']+" ["+location['geoRegion']+"]",
                attributes=[
                    "#defaultRouters",
                    "#"+location['locationCode'],
                    "#"+location['provider'],
                    "#"+location['geoRegion']
                ],
                provider="AWS",
                location_code=location['locationCode']
            )
            hosted_edge_routers += [er]
            print("INFO: Placed Edge Router in {major} ({location_name})".format(
                major=location['geoRegion'],
                location_name=location['locationName']
            ))

    for router_id in [r['id'] for r in hosted_edge_routers]:
        try:
            network.wait_for_status("PROVISIONED",id=router_id,type="edge-router",wait=999,progress=True)
        except:
            raise

    # create a simple global Edge Router Policy unless one exists with the same name
    ERPs = network.edge_router_policies()
    blanket_policy_name = "defaultRouters"
    if not blanket_policy_name in [erp['name'] for erp in ERPs]:
        try: network.create_edge_router_policy(name=blanket_policy_name,edge_router_attributes=["#defaultRouters"],endpoint_attributes=["#all"])
        except: raise

    endpoints = network.endpoints()
    clients = list()
    client1_name = "Desktop1"
    if not client1_name in [end['name'] for end in endpoints]:
        # create an Endpoint for the dialing device that will access Services
        client1 = network.create_endpoint(name=client1_name,attributes=["#workFromAnywhere"])
        print("INFO: created Endpoint \"{:s}\"".format(client1['name']))
    else:
        client1 = [end for end in endpoints if end['name'] == client1_name][0]
        print("INFO: found Endpoint \"{:s}\"".format(client1['name']))
    clients += [client1]

    client2_name = "Mobile1"
    if not client2_name in [end['name'] for end in endpoints]:
        # create an Endpoint for the dialing device that will access Services
        client2 = network.create_endpoint(name=client2_name,attributes=["#workFromAnywhere"])
        print("INFO: created Endpoint \"{:s}\"".format(client2['name']))
    else:
        client2 = [end for end in endpoints if end['name'] == client2_name][0]
        print("INFO: found Endpoint \"{:s}\"".format(client2['name']))
    clients += [client2]

    if args.client:
        client3_name = "Linux1"
        if not client3_name in [end['name'] for end in endpoints]:
            # create an Endpoint for the dialing device that will access Services
            client3 = network.create_endpoint(name=client3_name,attributes=["#workFromAnywhere"])
            print("INFO: created Endpoint \"{:s}\"".format(client3['name']))
        else:
            client3 = [end for end in endpoints if end['name'] == client3_name][0]
            print("INFO: found Endpoint \"{:s}\"".format(client3['name']))
        clients += [client3]

    exits = list()
    if args.private:
        exit1_name = "Exit1"
        if not exit1_name in [end['name'] for end in endpoints]:
            # create an Endpoint for the hosting device that will provide access to the server
            exit1 = network.create_endpoint(name=exit1_name,attributes=["#exits"])
            print("INFO: created Endpoint \"{:s}\"".format(exit1['name']))
        else:
            exit1 = [end for end in endpoints if end['name'] == exit1_name][0]
            print("INFO: found Endpoint \"{:s}\"".format(exit1['name']))
        exits += [exit1]

    # the demo containers have the demo working dir mounted on /netfoundry
    if os.access('/netfoundry', os.W_OK):
        token_path = '/netfoundry' # nosec
    else:
        token_path = str(Path.cwd())
    for end in clients+exits:
        token_file = token_path+'/'+end['name']+'.jwt'
        if end['jwt']:
            text = open(token_file, "wt")
            text.write(end['jwt'])
            text.close()
        else:
            if os.path.exists(token_file):
                print("DEBUG: cleaning up used OTT for enrolled Endpoint {end} from {path}".format(end=end['name'],path=token_file))
                os.remove(token_file)

    services = network.services()

    if args.private:
        # create Endpoint-hosted Services unless name exists
        service1_name = "Hello Service"
        if not service1_name in [svc['name'] for svc in services]:
            # traffic sent to hello.netfoundry:80 leaves Endpoint exit1 to server hello:3000
            service1 = network.create_service(
                name=service1_name,
                attributes=["#welcomeWagon"],
                client_host_name="hello.netfoundry",
                client_port="80",
                endpoints=[exit1['id']],
                server_host_name="hello",
                server_port="3000",
                server_protocol="TCP"
            )
            print("INFO: created Service \"{:s}\"".format(service1['name']))
        else:
            service1 = [svc for svc in services if svc['name'] == service1_name][0]
            print("INFO: found Service \"{:s}\"".format(service1['name']))

        service2_name = "REST Service"
        if not service2_name in [svc['name'] for svc in services]:
            # traffic sent to httpbin.netfoundry:80 leaves Endpoint exit1 to server httpbin:80
            service2 = network.create_service(
                name=service2_name,
                attributes=["#welcomeWagon"],
                client_host_name="httpbin.netfoundry",
                client_port="80",
                endpoints=[exit1['id']],
                server_host_name="httpbin",
                server_port="80",
                server_protocol="TCP"
            )
            print("INFO: created Service \"{:s}\"".format(service2['name']))
        else:
            service2 = [svc for svc in services if svc['name'] == service2_name][0]
            print("INFO: found Service \"{:s}\"".format(service2['name']))

    # Create router-hosted Services unless exists
    hosting_router = random.choice(hosted_edge_routers) # nosec

    service3_name = "Fireworks Service"
    if not service3_name in [svc['name'] for svc in services]:
        # traffic sent to fireworks.netfoundry:80 leaves Routers to 34.204.78.203:80
        service3 = network.create_service(
            name=service3_name,
            attributes=["#welcomeWagon"],
            client_host_name="fireworks.netfoundry",
            client_port="80",
            egress_router_id=hosting_router['id'],
            server_host_name="fireworks-load-balancer-1246256380.us-east-1.elb.amazonaws.com",
            server_port="80",
            server_protocol="TCP"
        )
        print("INFO: created Service \"{:s}\"".format(service3['name']))
    else:
        service3 = [svc for svc in services if svc['name'] == service3_name][0]
        print("INFO: found Service \"{:s}\"".format(service3['name']))

    service4_name = "Weather Service"
    if not service4_name in [svc['name'] for svc in services]:
        # traffic sent to weather.netfoundry:80 leaves Routers to wttr.in:80
        service4 = network.create_service(
            name=service4_name,
            attributes=["#welcomeWagon"],
            client_host_name="weather.netfoundry",
            client_port="80",
            egress_router_id=hosting_router['id'],
            server_host_name="wttr.in",
            server_port="80",
            server_protocol="TCP"
        )
        print("INFO: created Service \"{:s}\"".format(service4['name']))
    else:
        service4 = [svc for svc in services if svc['name'] == service4_name][0]
        print("INFO: found Service \"{:s}\"".format(service4['name']))

    # fireworks
    # heartbeat
    
    service5_name = "Echo Service"
    if not service5_name in [svc['name'] for svc in services]:
        # traffic sent to echo.netfoundry:80 leaves Routers to eth0.me:80
        service5 = network.create_service(
            name=service5_name,
            attributes=["#welcomeWagon"],
            client_host_name="echo.netfoundry",
            client_port="80",
            egress_router_id=hosting_router['id'],
            server_host_name="eth0.me",
            server_port="80",
            server_protocol="TCP"
        )
        print("INFO: created Service \"{:s}\"".format(service5['name']))
    else:
        service5 = [svc for svc in services if svc['name'] == service5_name][0]
        print("INFO: found Service \"{:s}\"".format(service5['name']))

    # create a customer-hosted ER unless exists
    customer_routers = network.edge_routers(only_customer=True)
    customer_router_name="Branch Exit Router 1"
    if not customer_router_name in [er['name'] for er in customer_routers]:
        customer_router = network.create_edge_router(
            name=customer_router_name,
            attributes=["#vmWareExitRouters"],
            tunneler_enabled=True
        )
    else:
        customer_router = [er for er in customer_routers if er['name'] == customer_router_name][0]
    # wait for customer router to be PROVISIONED so that registration will be available 
    try:
        network.wait_for_status("PROVISIONED",id=customer_router['id'],type="edge-router",wait=999,progress=True)
        customer_router_registration = network.rotate_edge_router_registration(id=customer_router['id'])
    except:
        raise
    print("INFO: Ready to register branch exit Edge Router {name} with key {key} (expires {expiry})".format(
        name=customer_router_name,
        key=customer_router_registration['registrationKey'],
        expiry=customer_router_registration['expiresAt'],
    ))


    # create unless exists
    app_wan1_name = "Welcome"
    app_wans = network.app_wans()
    if not app_wan1_name in [aw['name'] for aw in app_wans]:
        # workFromAnywhere may connect to welcomeWagon
        app_wan1 = network.create_app_wan(name=app_wan1_name,endpoint_attributes=["#workFromAnywhere"],service_attributes=["#welcomeWagon"])
        print("INFO: created AppWAN \"{:s}\"".format(app_wan1['name']))
    else:
        app_wan1 = [aw for aw in app_wans if aw['name'] == app_wan1_name][0]
        print("INFO: found AppWAN \"{:s}\"".format(app_wan1['name']))

    print("SUCCESS! The next step is to enroll one or more of your client Endpoints on some device(s) and visit one of the demo Service URLs described in the demo document ({doc})."
            "You may also log in to the web console ({nfconsole}) to play with your Network".format(doc="https://developer.netfoundry.io/v2/tools/#demos",nfconsole=network_group.nfconsole))
    for svc in network.services():
        print("* {name}:\thttp://{url}/".format(name=svc['name'],url=svc['model']['clientIngress']['host']))

def query_yes_no(question: str, default: str="no"):
    """Ask a yes/no question via input() and return their answer.

    "question" is a string that is presented to the user.
    "default" is the presumed answer if the user just hits <Enter>.
        It must be "yes" (the default), "no" or None (meaning
        an answer is required of the user).

    The "answer" return value is True for "yes" or False for "no".
    """
    valid = {"yes": True, "y": True, "ye": True,
            "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if default is not None and choice == '':
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            sys.stdout.write("Please respond with 'yes' or 'no' "
                            "(or 'y' or 'n').\n")

if __name__ == '__main__':
    main()


"""Print the version string."""
from ._version import get_versions
try:
    print("v"+get_versions()['version'] )
except:
    # package is not installed
    raise Exception("ERROR: could not do get_version() from versioneer.")




import versioneer
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='netfoundry',
    py_modules=['netfoundry'],
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://developer.netfoundry.io/guides/python/',
    description='Interface to the NetFoundry network-as-code orchestration Platform',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license='MIT',
    author='Kenneth Bingham',
    author_email='support@netfoundry.io',
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'requests >= 2.25.1',
        'pysocks >= 1.7.1',
        'pyjwt >= 2.0.0',
        'inflect >= 5.3'
    ]
)

